import Combine
import Foundation

public enum Bus {
  private static let e = PassthroughSubject<(key: String, value: Any), Never>()

  public static var events: AnyPublisher<(key: String, value: Any), Never> {
    e.eraseToAnyPublisher()
  }
}

public extension Bus {
  enum Option {
    case async
  }
}

public extension Bus {
  static func receive<T>(
    _ keys: Set<String>,
    _ handler: @escaping ((String, T) -> Void),
    opt: [Option] = [],
    sub subscriptions: inout [AnyCancellable]
  ) {
    let isAsync = opt.contains(.async)

    // Async.
    if isAsync {
      Self.events
        .compactMap { convertKeyValue(keys, $0) }
        .receive(on: DispatchQueue.main)
        .sink { v in handler(v.0, v.1) }
        .store(in: &subscriptions)
    }

    // Async.
    if !isAsync {
      Self.events
        .compactMap { convertKeyValue(keys, $0) }
        .sink { v in handler(v.0, v.1) }
        .store(in: &subscriptions)
    }
  }

  static func send<T>(
    _ key: String,
    _ node: AnyPublisher<T, Never>,
    opt: [Option] = [],
    sub subscriptions: inout [AnyCancellable]
  ) {
    let isAsync = opt.contains(.async)

    // Async.
    if isAsync {
      node
        .receive(on: DispatchQueue.main)
        .sink { v in Self.e.send((key, v)) }
        .store(in: &subscriptions)
    }

    // Sync.
    if !isAsync {
      node
        .sink { v in Self.e.send((key, v)) }
        .store(in: &subscriptions)
    }
  }
  
  static func send(_ key: String, _ value: Any) {
    Self.e.send((key, value))
  }
}

public extension Bus {
  static func process<Src, Dst>(
    _ keysIn: Set<String>,
    _ keyOut: String,
    _ handler: @escaping ((Src) -> Dst?),
    opt: [Option] = [],
    sub subscriptions: inout [AnyCancellable]
  ) {
    let isAsync = opt.contains(.async)

    // Async.
    if isAsync {
      Self.events
      .compactMap { processKeysValue($0, keysIn, handler) }
      .receive(on: DispatchQueue.main)
      .sink { vOut in Self.e.send((keyOut, vOut)) }
      .store(in: &subscriptions)
    }

    // Sync.
    if !isAsync {
      Self.events
      .compactMap { processKeysValue($0, keysIn, handler) }
      .sink { vOut in Self.e.send((keyOut, vOut)) }
      .store(in: &subscriptions)
    }
  }
}
